<?php
( defined( 'ABSPATH' ) && is_admin() && is_user_logged_in() && ( current_user_can( 'manage_options' ) || current_user_can( 'corona_test_results_manage' ) ) ) or die;

global $corona_test_results_cfg;

/**
 * Generate a random number
 * @see https://stackoverflow.com/questions/1846202/php-how-to-generate-a-random-unique-alphanumeric-string-for-use-in-a-secret-l
 */
function corona_test_results_rand_secure($min, $max) {
    $range = $max - $min;
    if ($range < 1) return $min; // not so random...
    $log = ceil(log($range, 2));
    $bytes = (int) ($log / 8) + 1; // length in bytes
    $bits = (int) $log + 1; // length in bits
    $filter = (int) (1 << $bits) - 1; // set all lower bits to 1
    do {
        $rnd = hexdec(bin2hex(openssl_random_pseudo_bytes($bytes)));
        $rnd = $rnd & $filter; // discard irrelevant bits
    } while ($rnd > $range);
    return $min + $rnd;
}

/**
 * generate a random code of characters in our character table
 */
function corona_test_results_get_random_code($length)
{
    $codeAlphabet = corona_test_results_get_code_chars();

    $token = "";
    $max = strlen($codeAlphabet);

    for ($i=0; $i < $length; $i++) {
        $token .= $codeAlphabet[corona_test_results_rand_secure(0, $max-1)];
    }

    return $token;
}

/**
 * Checks if the table used by the plugin exists, and otherwise creates it.
 */
function corona_test_results_conditionally_create_table() {
    require_once( ABSPATH . 'wp-admin/includes/upgrade.php' );

    $tableName = corona_test_results_get_table_name();
    $main_sql_create = "CREATE TABLE $tableName (
        code varchar(" . CORONA_TEST_RESULTS_CODE_COLUMN_LENGTH . ") NOT NULL,
        status tinyint(3) UNSIGNED NOT NULL DEFAULT 0,
        created_at timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
		trash tinyint(3) UNSIGNED NOT NULL DEFAULT 0,
        PRIMARY KEY  (code)
    );";

    // $wpdb->show_errors = true;
	return dbDelta( $main_sql_create );
}

/**
 * Register our admin pages in the menu and triggers enqueueing our styles and scripts
 */
function corona_test_results_add_test_results_menu() {
	$page_title = __('Test Result Assignation', 'corona-test-results');
	$menu_title = __('Test Results', 'corona-test-results');
	$capability = 'manage_options';
	$menu_slug  = 'corona-test-results';
	$function   = 'corona_test_results_adminpage';
	$icon_url   = 'data:image/svg+xml;base64,' . base64_encode( file_get_contents( plugin_dir_path( __FILE__  ) . 'assets/corona-test-results-admin-icon.svg' ) );
	$position   = 4;

	add_menu_page( $page_title, $menu_title, $capability, $menu_slug, $function, $icon_url, $position );

	add_submenu_page(
		'corona-test-results',
		__('Test Result Assignation', 'corona-test-results'),
		__('Assign results', 'corona-test-results'),
		'manage_options',
		'corona-test-results',
		'corona_test_results_adminpage'
	);
	add_submenu_page(
		'corona-test-results',
		__('Register test', 'corona-test-results'),
		__('Register test', 'corona-test-results'),
		'manage_options',
		'corona-test-results-register',
		'corona_test_results_adminpage_register'
	);

	add_action( 'admin_enqueue_scripts', 'corona_test_results_assets' );
}
add_action( 'admin_menu', 'corona_test_results_add_test_results_menu' );

/**
 * Returns the internal slugs for our different result pages.
 * This is also used to iterate over all the different page types to filter or generate content.
 */
function corona_test_results_get_result_pages_slugs() {
	return ['result_retrieval', 'result_pending', 'result_positive', 'result_negative'];
}

/**
 * Returns a human-readable and translatable label from the internal slugs of our different result pages
 */
function corona_test_results_get_page_name( $page_slug ) {
	$name = null;

	switch ($page_slug) {
		case 'result_retrieval':
			$name = __( 'Test Results Retrieval Page', 'corona-test-results' );
			break;
		case 'result_pending':
		case 'result_positive':
		case 'result_negative':
			// translators: %s: test result status
			$name = sprintf( __( 'Result Page Content: %s', 'corona-test-results' ), corona_test_results_get_page_state( $page_slug ) );
			break;
	};

	return $name;
}

/**
 * returns the slug prefix of admin pages, because hooks are using the translated value
 */
function corona_test_results_admin_page_slug($page = 'settings') {
	return sanitize_title(__('Test Results', 'corona-test-results')) . '_page_corona-test-results-' . $page;
}

/**
 * Enque all our styles and scripts when and where needed
 */
function corona_test_results_assets($hook) {
	if( $hook === 'toplevel_page_corona-test-results' || $hook === corona_test_results_admin_page_slug('register') ) {
		wp_register_script( 'corona-test-results-pdfmake', plugin_dir_url( __FILE__ ) . '/assets/js/pdfmake/pdfmake.min.js', [], filemtime( plugin_dir_path( __FILE__ ) . 'assets/js/pdfmake/pdfmake.min.js' ) );
		wp_enqueue_script( 'corona-test-results-pdfmake' );
		wp_register_script( 'corona-test-results-pdfmake-fonts', plugin_dir_url( __FILE__ ) . '/assets/js/pdfmake/vfs_fonts_custom.min.js', [], filemtime( plugin_dir_path( __FILE__ ) . 'assets/js/pdfmake/vfs_fonts_custom.min.js' ) );
		wp_enqueue_script( 'corona-test-results-pdfmake-fonts' );

		wp_register_script( 'corona-test-results', plugin_dir_url( __FILE__ ) . '/assets/js/corona-test-results.js', [], filemtime( plugin_dir_path( __FILE__ ) . 'assets/js/corona-test-results.js' ) );
		wp_enqueue_script( 'corona-test-results' );

		wp_localize_script( 'corona-test-results', 'corona_test_results',
			array(
				'resultsBaseUrl' => corona_test_results_get_page_url(),
				'premiumFeatureNotice' => __( 'This feature is only available in the Premium version of the plugin.', 'corona-test-results' ),
				'rows_loading' => __( 'Loading codes…', 'corona-test-results' ),
				'rows_loading_failed' => __( 'Failed fetching additional table rows!', 'corona-test-results' ),
				'columnHeaders' => array(
					__('Code', 'corona-test-results'),
					__('Status', 'corona-test-results'),
					__('Test Date', 'corona-test-results')
				),
				'cfg' => array_filter(
					corona_test_results_get_options(),
					function ($key) {
						return strpos($key, 'template_') === 0;
					},
					ARRAY_FILTER_USE_KEY
				),
				'documentTexts' => array(
					'onsite_notice' => __('[for on-site safekeeping!]', 'corona-test-results'),
					'labels' => array(
						'surname' => __('Surname', 'corona-test-results'),
						'firstname' => __('First Name', 'corona-test-results'),
						'dateofbirth' => __('Date of Birth', 'corona-test-results'),
						'testdate' => __('Test Date', 'corona-test-results')
					),
					'date_i18n' => date_i18n( get_option( 'date_format' ) ),
					'powered_by' => sprintf(__('Online test results powered by %s', 'corona-test-results'), '<strong>48DESIGN</strong>.com | VISIONS. PROGRAMMING. DESIGN')
				)
			)
		);
	} else if ( $hook === corona_test_results_admin_page_slug('settings') ) {
		wp_register_script( 'corona-test-results-settings', plugin_dir_url( __FILE__ ) . '/assets/js/corona-test-results-settings.js', [], filemtime( plugin_dir_path( __FILE__ ) . 'assets/js/corona-test-results-settings.js' ) );
		wp_enqueue_script( 'corona-test-results-settings' );
	}
}

function corona_test_results_get_entries( $stati, $offset = 0 ) {
	global $wpdb;

	$tableName = corona_test_results_get_table_name();

	$limitPhrase = '';

	if ( $offset > 0 ) {
		$limitPhrase = ' LIMIT ' . intval($offset) . ', 18446744073709551615';
	}

	$rowsSQL = $wpdb->prepare("SELECT `code`, `status`, `created_at` FROM `$tableName` WHERE `status` IN(" . implode(', ', array_fill( 0, count( $stati ), '%d' ) ) . ") ORDER BY `created_at` DESC", $stati ) . $limitPhrase;
    $rowsQuery = $wpdb->get_results ( $rowsSQL );

	return $rowsQuery;
}

/**
 * Outputs the content of the test assignation admin page
 */
function corona_test_results_adminpage() {
    global $wpdb, $corona_test_results_cfg;

    if( !current_user_can( 'manage_options' ) ) {
        wp_die(__( 'Sorry, you are not allowed to access this page.' ));
    }

    corona_test_results_conditionally_create_table();
    $tableName = corona_test_results_get_table_name();

	$page_param = '';
	$status_param = '';

	if ( isset( $_GET['page'] ) ) {
		$page_param = sanitize_key( $_GET['page'] );
	}

	if ( isset( $_GET['status'] ) ) {
		$status_param = sanitize_text_field( $_GET['status'] );
	}

    $filterStatus = ( is_numeric($status_param) ? (int)$status_param : $status_param ) ?: 0;

    $currentAttrs = ' class="current" aria-current="page"';

    $statusCount = array_fill(0,3,0);

    $statusCountQuery = $wpdb->get_results ( "SELECT status, COUNT(*) as count FROM `$tableName` GROUP BY `status` ORDER BY count DESC" );
    foreach ($statusCountQuery as $row) {
        $statusCount[$row->status] = $row->count;
    }

	// insert 'Evaluated' into second tab position
	$insertPosition = 1;
	$statusTabs = corona_test_results_get_states();
	$statusTabs = array_slice( $statusTabs, 0, $insertPosition, true )
					+ array( '1|2' => __( 'Evaluated' , 'corona-test-results' ) )
					+ array_slice( $statusTabs, $insertPosition, count( $statusTabs ) - 1, true );
	$statusCount['1|2'] = $statusCount[1] + $statusCount[2];

	?>
	<div class="wrap" id="corona_test_results_assign">
		<h1>
			<?php esc_html_e('Test Result Assignation', 'corona-test-results'); ?>
		</h1>
		<?php corona_test_results_check_review_nag(); ?>
        <ul class="subsubsub">
        <?php
            $n=0;
            foreach($statusTabs as $status => $tab) {
                $n++;
                ?>
            <li><a href="?page=<?php echo urlencode( $page_param ) ?>&status=<?php echo urlencode( $status ) ?>"<?php echo $filterStatus === $status ? $currentAttrs : '' ?>>
                <?php echo esc_html_e($tab, 'corona-test-results') ?> <span class="count">(<?php echo $statusCount[$status] ?>)</span>
            </a><?php echo $n < count($statusTabs) ? ' |' : ''?></li>
                <?php
            }
        ?>
        </ul>

    <div class="tablenav top">
		<div class="alignleft">
			<label for="corona_test_results_search"><?php _e('Filter codes:', 'corona-test-results'); ?></label> <input type="search" id="corona_test_results_search" disabled>
			<span class="divider">|</span> <button type="button" id="corona_test_results_export" class="button button-secondary" disabled><?php _e('Export to CSV', 'corona-test-results'); ?></button>
		</div>

		<?php
			$itemsPerPage = $corona_test_results_cfg['assignation_entries_per_page'];
			$itemCount = $statusCount[$filterStatus];
			$pageCount = (int)ceil($itemCount / $itemsPerPage);
			$currentPage = 1;

			if ( isset( $_GET['paged'] ) ) {
				$currentPage = (int)$_GET['paged'];

				if ( $currentPage < 1 ) {
					$currentPage = 1;
				} else if ( $currentPage > $pageCount ) {
					$currentPage = $pageCount;
				}
			}

			if ($itemCount > $itemsPerPage) {

		?>
		<div class="tablenav-pages">
			<input type="hidden" id="pagination-items-per-page" value="<?php echo $itemsPerPage; ?>" />
			<span class="displaying-num"><?php echo sprintf( '%d ' . vt_helper__default_i18n__( 'items' ), $itemCount ); ?></span>
			<button type="button" class="first-page button" disabled><span class="screen-reader-text"><?php vt_helper__default_i18n_e( 'First page' ); ?></span><span aria-hidden="true">«</span></button>
			<button type="button" class="prev-page button" disabled><span class="screen-reader-text"><?php vt_helper__default_i18n_e( 'Previous page' ); ?></span><span aria-hidden="true">‹</span></button>
			<span id="table-paging" class="paging-input"><label for="current-page-selector" class="screen-reader-text"><?php vt_helper__default_i18n_e( 'Current Page' ); ?></label><input class="current-page" id="current-page-selector" type="text" name="paged" value="<?php echo esc_attr($currentPage); ?>" size="1" aria-describedby="table-paging"><span class="tablenav-paging-text"> <?php vt_helper__default_i18n_e( 'of' );  ?> <span class="total-pages"><?php echo $pageCount ?></span></span></span>
			<button type="button" class="next-page button"><span class="screen-reader-text"><?php vt_helper__default_i18n_e( 'Next page' ); ?></span><span aria-hidden="true">›</span></button>
			<button type="button" class="last-page button"<?php if( $pageCount < 3) { echo ' disabled'; } ?>><span class="screen-reader-text"><?php vt_helper__default_i18n_e( 'Last page' ); ?></span><span aria-hidden="true">»</span></button>
		</div>
		<?php
		}
		?>
    </div>

	<style>
		#corona_test_results_table code {
			background-color: transparent;
			letter-spacing: 0.1em;
			font-size: 1.1rem;
			display: inline-block;
		}

		#corona_test_results_table th,
		#corona_test_results_table td {
			vertical-align: middle;
		}

		@media only screen and (max-width: 960px) {
			.tablenav {
				text-align: center;
			}

			.tablenav .alignleft {
				width: 100%;
			}

			.tablenav .alignleft label {
				display: block;
			}

			.tablenav .alignleft .divider {
				display: block;
				overflow: hidden;
				height: 5px;
			}

			#corona_test_results_export {
				margin-bottom: 2em;
			}
		}
	</style>
    <form method="POST" action="<?php echo admin_url( 'admin-post.php' ) ?>" id="corona-test-results-form" autocomplete="off">
    <input type="hidden" name="action" value="corona_test_results_assign">
    <input type="hidden" name="from_status" value="<?php echo esc_attr( $filterStatus ) ?>">
    <input type="hidden" name="paged" value="<?php echo esc_attr( isset($_GET['paged']) ? (int)$_GET['paged'] : '1' ); ?>">
    <input type="hidden" name="corona_test_results_nonce" value="<?php echo wp_create_nonce( 'corona_test_results_assign_nonce' );  ?>" />
    <table id="corona_test_results_table" class="wp-list-table widefat fixed striped table-view-list pages">
        <thead>
        <tr>
            <!-- <td id="cb" class="manage-column column-cb check-column"><label class="screen-reader-text" for="cb-select-all-1">Alle auswählen</label><input id="cb-select-all-1" type="checkbox"></td> -->
            <th scope="col" class="manage-column"><?php _e('Code', 'corona-test-results'); ?></th>
            <th scope="col" class="manage-column"><?php _e('Status', 'corona-test-results'); ?></th>
            <th scope="col" class="manage-column"><?php _e('Test Date', 'corona-test-results'); ?></th>
        </tr>
        </thead>
        <tbody>
    <?php
	$stati = explode( '|', $filterStatus );

	$rowsQuery = corona_test_results_get_entries( $stati );

	$hasHiddenRows = false;

    if (empty($rowsQuery)) {
        print '<tr><td colspan="3">' . sprintf(esc_html__('There are no codes with a status of &quot;%s&quot; yet', 'corona-test-results'), $statusTabs[$filterStatus]) . '</td></tr>';
    } else {
		$rowNum = 0;
        foreach ($rowsQuery as $row) {
			$rowNum++;
			if ( $rowNum > $itemsPerPage ) {
				$hasHiddenRows = true;
				break;
			}
			$displayTime = date_i18n( get_option( 'date_format' ) . ', ' . get_option( 'time_format' ), strtotime(get_date_from_gmt($row->created_at, 'Y-m-d H:i:s')) );
			$timestamp = get_date_from_gmt($row->created_at, 'Y-m-d H:i:s');
        ?>
        <tr>
            <!-- <td></td> -->
            <td><code><?php echo esc_html($row->code) ?></code></td>
            <td>
            <select name="corona_test_results_result_status[<?php echo esc_attr( $row->code ) ?>]">
            <?php
                foreach(corona_test_results_get_states() as $status => $label) {
                    ?>
                    <option value="<?php echo esc_attr( $status ) ?>"<?php echo (int)$row->status === (int)$status ? ' selected' : '' ?>><?php echo esc_html( $label ) ?></option>
                    <?php
                }
            ?>
            </select>
            </td>
            <td data-datetime="<?php echo $timestamp; ?>"><?php echo $displayTime; ?></td>
        </tr>
        <?php
        }
    }
    ?>
        </tbody>
    </table>
	<?php
	if ( $hasHiddenRows ) {
	?>
	<script>window.ctr_fetch_hidden_rows = <?php echo json_encode($stati); ?></script>
	<?php
	}
	?>
    <p class="submit">
        <input type="submit" id="submit" class="button button-primary" value="<?php vt_helper__default_i18n_e('Save Changes'); ?>">
    </p>
    </form>
	</div>
	<?php
}

/**
 * update the status of one or multiple codes
 */
function corona_test_results_assign_response() {
    global $wpdb;

    if( !current_user_can( 'manage_options' ) || !isset( $_POST['corona_test_results_nonce'] ) || !wp_verify_nonce( $_POST['corona_test_results_nonce'], 'corona_test_results_assign_nonce') ) {
        wp_die(__( 'Sorry, you are not allowed to access this page.' ));
    }

    if ( isset( $_POST['corona_test_results_result_status'] ) && count( $_POST['corona_test_results_result_status'] ) ) {
		if ( !empty( $_POST['corona_test_results_result_status_hidden'] )) {
			$additionalStateChanges = json_decode( stripslashes( $_POST['corona_test_results_result_status_hidden'] ), true );
			if ( !empty( $additionalStateChanges ) && is_array( $additionalStateChanges ) ) {
				$_POST['corona_test_results_result_status'] = $_POST['corona_test_results_result_status'] + $additionalStateChanges;
			}
		}
		$tableName = corona_test_results_get_table_name();

		$update_string = str_repeat( "WHEN `code` = %s THEN %d \n", count( $_POST['corona_test_results_result_status'] ) );
		$update_stmt = "UPDATE `$tableName` SET
			`status` = (CASE
							$update_string
							ELSE `status`
						END
						)";

		$values = array();
        foreach ( $_POST['corona_test_results_result_status'] as $code=>$newStatus ) {
			$values[] = strtoupper( sanitize_key( $code ) );
			$values[] = (int)$newStatus;
		}

		// $wpdb->show_errors = true;
		$wpdb->query( $wpdb->prepare( $update_stmt, $values ) );
		// echo $wpdb->last_query;
		// exit;
    }

    wp_safe_redirect(
        admin_url(
			'/admin.php?page=corona-test-results&status='
				. urlencode( sanitize_text_field( $_POST['from_status'] ) )
				. ( isset( $_POST['paged'] ) && (int)$_POST['paged'] > 1 ? '&paged=' . (int)$_POST['paged'] : '' )
		)
    );
    exit;
}
add_action( 'admin_post_corona_test_results_assign', 'corona_test_results_assign_response');

/**
 * returns the plugin's support email address
 */
function corona_test_results_get_support_mail() {
	return 'wordpress@48design.com';
}

/**
 * returns an HTML link to the plugin's support email address
 */
function corona_test_results_get_support_mail_link( $text = null ) {
	return '<a href="mailto:' . esc_attr( corona_test_results_get_support_mail() ) . '">' . ( $text ? $text : __( 'Support', 'corona-test-results' ) ). '</a>';
}
/**
 * returns an HTML link to the plugin's review page
 */
function corona_test_results_get_review_link( $prefix = '' ) {
	// translators: used as link text in the review notice ("... please take some time to _rate and review_ the plugin ...") as well as standalone on the plugin overview page and should fit both cases gramatically
	return "<a href='https://wordpress.org/plugins/corona-test-results/#reviews' target='_blank' rel='noopener'>" . $prefix . __( "rate and review", 'corona-test-results' ) . "</a>";
}

/**
 * returns the URL to the plugin installation page
 */
function corona_test_results_get_add_plugins_url( $show_upload_tab = true ) {
	$baseUrl = get_admin_url() . 'plugin-install.php';
	$settings_url = esc_url( $show_upload_tab ? add_query_arg(
		'tab',
		'upload',
		$baseUrl
	) : $baseUrl );

	return $settings_url;
}

/**
 * returns the URL of the plugin's settings page
 */
function corona_test_results_get_settings_url() {
	$settings_url = esc_url( add_query_arg(
		'page',
		'corona-test-results-settings',
		get_admin_url() . 'admin.php'
	) );

	return $settings_url;
}

/**
 * returns an HTML link to the settings page
 */
function corona_test_results_get_settings_link() {
	$settings_url = corona_test_results_get_settings_url();

	$settings_page_link = '<a href="' . $settings_url . '">' . esc_html__('Test Results Settings Page', 'corona-test-results') . '</a>';

	return $settings_page_link;
}

/**
 * Check whether all the required result pages are set
 */
function corona_test_results_check_all_pages_set() {
	return corona_test_results_get_page_id( 'result_retrieval' );
}

/**
 * show rate and review nag notice
 */
function corona_test_results_check_review_nag() {
	if (
		!get_option('dismissed_corona_test_results_rating_nag', false )
		&& corona_test_results_check_review_nag_timestamp()
	) {
		echo "<div class='updated notice notice-info is-dismissible' data-dismiss-id='rating_nag'><p>"
		// translators: %s "rate and review" linking to the plugin review page on WordPress.com
		. nl2br(
			sprintf(
				__( "You have been using the Corona Test Results plugin for a while now. If you like it or have some feedback to share, please take a minute to %s the plugin.", 'corona-test-results' ),
				corona_test_results_get_review_link()
			)
		)
		. "</p></div>";
	}
}

function corona_test_results_check_review_nag_timestamp() {
	return (
		( $activationtime = get_option( 'corona_test_results_activationtime' ) )
		&& ( time() - $activationtime ) / 60 / 60 / 24 >= 14 // plugin activation has been at least 14 days ago
	);
}

/**
 * Outputs the content for the test registration admin page
 */
function corona_test_results_adminpage_register() {
    $options = corona_test_results_get_options();

	$premium_string = '<strong><em>Premium</em></strong>';
	$upsell_message = nl2br(
		esc_html__( "You are using the free version of the plugin.", 'corona-test-results')
		. ' ' . sprintf(
			// translators: %s: (1) formatted string "Premium", (2,3) opening and closing link tags
			esc_html__( "%sGet the %s version%s in order to customize code generation, documents and result page contents.", 'corona-test-results' ),
			'<a href="' . esc_url( corona_test_results_premium_shop_url() ) . '" target="_blank">',
			$premium_string,
			'</a>'
		)
	);
	echo "<div class='notice notice-warning'><p>$upsell_message</p></div>";

	if (preg_match('~MSIE|Internet Explorer~i', $_SERVER['HTTP_USER_AGENT']) || preg_match('~Trident/7.0(; Touch)?; rv:11.0~',$_SERVER['HTTP_USER_AGENT'])) {
		echo "<div class='notice notice-warning'><p><strong>"
			. nl2br( __( "You are using an outdated browser that does not perform very well, especially during batch PDF generation. If you experience long processing times, freezing of the browser window or failing PDF generation, please switch to a modern browser.", 'corona-test-results' ) )
		. "</strong></p></div>";
	}

	if( !current_user_can( 'manage_options' ) ) {
		wp_die(__( 'Sorry, you are not allowed to access this page.' ));
	}

	$settings_page_link = corona_test_results_get_settings_link();

	$general_error = null;
	if (!corona_test_results_check_all_pages_set()) {
		// translators: %s: "Test Results Settings Page" with link
		$general_error = sprintf( esc_html__('Not all result pages are set. Please check the %s and select or create the pages for the result retrieval form and result contents.', 'corona-test-results'), $settings_page_link );
	} else if (corona_test_results_get_page( 'result_retrieval' )->post_status !== 'publish') {
		$general_error = sprintf( esc_html__('The test results retrieval page has not yet been published and will not be accessible to visitors of the provided URL.', 'corona-test-results'), $settings_page_link );
	}
	?>
	<style>
		.wrap + .wrap {
			margin-top: 2em;
		}

		#submit[disabled]::after,
		#generate-pdf[disabled]::after,
		#generate-batch[disabled]::after,
		#batch-export-csv[disabled]::after,
		#batch-export-pdf[disabled]::after {
			display: inline-block;
			content: "";
			margin-left: 0.5em;
			border: dotted 2px;
			width: 1em;
			height: 1em;
			border-radius: 50%;
			vertical-align: middle;
			margin-top: -0.2em;
			animation: busy-rotate 2s linear infinite;
		}

		@keyframes busy-rotate {
			0% {
				transform: rotate(0deg);
			}

			100% {
				transform: rotate(360deg);
			}
		}
	</style>
	<div id="message" class="error" style="display: <?php print !empty($general_error) ? 'block' : 'none' ?>">
		<p class="vt-error-ajax" style="display: none"><?php esc_html_e('An error occurred during code generation. Please try again.', 'corona-test-results'); ?></p>
		<p class="vt-error-codegen" style="display: none"><?php
			// translators: %s: "Test Results Settings Page" with link
			$settingsPageHint = sprintf( esc_html__('Please go to the %s and either increase the code length or reduce your block list.', 'corona-test-results'), $settings_page_link );
			print esc_html__('No valid code could be generated after multiple tries.')
				. ' ' . $settingsPageHint;
		?></p>
		<p class="vt-error-codegen-batch" style="display: none"><?php
			print esc_html__('Not all codes could be generated after multiple tries.')
				. ' ' . $settingsPageHint;
		?></p>
		<p class="vt-error-general" style="display: <?php print !empty($general_error) ? 'block' : 'none' ?>"><?php print $general_error; ?></p>
	</div>
	<div class="wrap" id="corona_test_results_register">
		<h1>
			<?php esc_html_e('Test Registration', 'corona-test-results'); ?>
		</h1>
		<?php corona_test_results_check_review_nag(); ?>
		<table class="form-table" role="presentation" id="corona-test-results-form">
			<tbody>
			<?php
				if ( !empty( $options['template_customfield'] ) ) {
			?>
			<tr>
				<th scope="row"><label for="customfield"><?php echo esc_html( $options['template_customfield'] ) ?></label></th>
				<td><input name="customfield" type="text" id="customfield" class="regular-text" autocomplete="off"></td>
			</tr>
			<?php
				}
			?>
			<tr>
				<th scope="row"><label for="surname"><?php esc_html_e('Surname', 'corona-test-results') ?></label></th>
				<td><input name="surname" type="text" id="surname" class="regular-text" autocomplete="off"></td>
			</tr>
			<tr>
				<th scope="row"><label for="firstname"><?php esc_html_e('First Name', 'corona-test-results') ?></label></th>
				<td><input name="firstname" type="text" id="firstname" class="regular-text" autocomplete="off"></td>
			</tr>
			<tr>
				<th scope="row"><label for="dateofbirth"><?php esc_html_e('Date of Birth', 'corona-test-results') ?></label></th>
				<td><input name="dateofbirth" type="date" id="dateofbirth" class="" autocomplete="off"></td>
			</tr>
			<tr>
				<th scope="row"><label for="test_result_code"><?php esc_html_e('Code for retrieval', 'corona-test-results') ?></label></th>
				<td><input name="test_result_code" type="text" id="test_result_code" class="regular-text" autocomplete="off" readonly placeholder="<?php esc_attr_e('(generated automatically)', 'corona-test-results') ?>"></td>
			</tr>
			</tbody>
		</table>

		<p class="submit">
			<button type="button" id="submit" class="button button-primary"><?php esc_html_e('Generate code and PDF', 'corona-test-results') ?></button>
			<button type="button" id="generate-pdf" class="button button-primary" style="display: none"><?php esc_html_e('Regenerate PDF', 'corona-test-results') ?></button>
			<input type="button" id="reset-form" class="button" value="<?php esc_attr_e('Register a new test', 'corona-test-results') ?>" style="display: none">
		</p>
		<p id="privacy-hint" class="description"><?php echo nl2br( __("The personal data is never sent to or stored on the server. Generation of the document is done completely in your browser.\nImmediately save or print out the document, as it will not be accessible in your browser at a later time.", 'corona-test-results') ); ?></p>
		<p id="popup-hint" class="description" style="display: none"><?php _e('If the generated document does not open, set your browser to allow popups for this website.', 'corona-test-results'); ?></p>
	</div>

	<div class="wrap">
		<h1><?php _e( 'Batch generation', 'corona-test-results' ); ?></h1>
		<p class="description"><?php _e( 'Generate multiple codes and documents, e.g. if there\'s no printer at the testing location, or you make home visits for testing.', 'corona-test-results' ) ?>
		<br>
		<?php
			echo "<strong>" . esc_html__('Note:', 'corona-test-results')
			. "</strong> " . esc_html__( 'PDF generation can take quite some time or even fail for more than a few hundreds of codes. If you plan to export codes to PDF instead of CSV, use a reasonably sized batch size of about 250 and do multiple runs if you need a larger number of codes.', 'corona-test-results' );
		?>
		</p>

		<p><?php
			// translators: %s: input field for numeric value
			echo sprintf( __( 'Generate %s codes', 'corona-test-results'), '<input type="number" class="small-text" min="2" max="9999" value="2" maxlen="4" id="batch-count" />');
		?></p>

		<p class="submit">
			<button type="button" id="generate-batch" class="button button-primary"><?php esc_html_e('Generate', 'corona-test-results') ?></button>
		</p>

		<p id="batch-actions" style="display: none;">
			<?php
			_e( 'Last generated batch:', 'corona-test-results' );
			?>
			<button type="button" id="batch-export-csv" class="button"><?php esc_html_e('Export to CSV', 'corona-test-results') ?></button>
			<button type="button" id="batch-export-pdf" class="button"><?php esc_html_e('Export to PDF', 'corona-test-results') ?></button>
		</p>
	</div>
	<?php
}

/**
 * returns a list of words that could be offending or otherwise undesirable in the generated codes,
 * including cases where numbers might look like a letter
 */
function corona_test_results_get_badword_list() {
	$blocklist = array(
		'123','18HH','18NS','18SS','4USCHW1TZ','88NS','88SS',
		'ABC','AFD','ARSCH','ARSE','ASDF','AUSCHW1TZ','AUSCHWITZ',
		'B4LLS','BALLS','B4ST4RD','B4STARD','BAST4RD','BASTARD','BL0WJ0B','BLAS','BLOWJ0B','BLOWJOB','BUMS','BUMS3N','BUMSEN','BUTT',
		'C0CK','CDU','COCK','CSU',
		'D00F','D0OF','D4CH4U','D4CHAU','DACH4U','DACHAU','DICK','DO0F','DOEDEL','DOOF','DRECK','DUMB','DUMM',
		'EKEL',
		'F1CK','FDP','FICK','FOTZE','FUCK',
		'G03BB3L','G03BBEL','G0EBB3L','G0EBBEL','G3N0Z1D','GEBER','GEN0Z1D','GEN0ZID','GENOZ1D','GENOZID','GOEBB3L','GOEBBEL',
		'HAESSL1CH','HAESSLICH','H0D3N','H0DEN','H1TL3R','H1TLER','H3IL','HEIL','HH18','HITL3R','HITLER','HOD3N','HODEN','HURE',
		'IDI0T','IDIOT',
		'JEW','JUDE',
		'K0TZ','K4CK','KACK','KOTZ',
		'MORD','MUERT','N1GG32',
		'N1GG3R','N1GGE2','N1GGER','N3G3R','N3GER','N4ZI','NAZI','NEG3R','NEGER','NIGG3R','NPD','NS18','NS88','NSAH','NSD4P','NSDAP','NUTTE',
		'P00P','P0OP','P0RN','P1MM3L','P1MMEL','P3N1S','P3NIS','PEN1S','PENIS','PENNER','PILLER','PIMM3L','PIMMEL','PIRAT','PO0P','POOP','POPEL','POPP','PORN','PR1CK','PRICK','PULLER','PUSSY','PUSSIE','PUSS1E',
		'QWERT',
		'S13G','S1EG','SACK','SAU','SCH31D3','SCH3ID3','SCH3IDE','SCHE1D3','SCHEIDE','SCHEIS','SCHWEIN','SEX','SH1T','SHIT','SI3G','SIEG','SS18','SS88','STERB','STERBEN','STIRB','SUCK',
		'T0SS3R','T0SSER','TITTEN','TOD','TOSS3R','TOSSER','TOT',
		'V1RG1N','V1RGIN','V4G1N4','V4GIN4','V4GINA','VAG1NA','VAGIN4','VAGINA','VERECK','VERRECK','VIRG1N','VIRGIN','VOTZE',
		'WANK','WANK3R','WANKER','WH0R3','WH0RE','WHOR3','WHORE','WICHS',
		'ZECKE'
	);

    $options = get_option( 'corona_test_results_options' );

	if (is_array($options) && isset($options['code_blocklist'])) {
		$blockStrings = preg_split("/[\n,\s]/", $options['code_blocklist']);
		$blockStrings = array_map('trim', $blockStrings);
		$blockStrings = array_filter($blockStrings, function($value) { return !is_null($value) && $value !== ''; });
		$blockStrings = array_map('strtoupper', $blockStrings);

		$blocklist = array_merge($blocklist, $blockStrings);
	}

	return $blocklist;
}

/**
 * check a generated code against our badwords list
 * and for easy-to-guess combinations
 */
function corona_test_results_code_hasBadWord($code) {
    // same character repeated more than 3 times
    if (preg_match("/(.)\\1{2,}/", $code)) {
        return true;
    }

    // same string repeated more than once
    if (preg_match("/(.+)\\1{1,}/", $code)) {
        return true;
    }

    $blockedWords = corona_test_results_get_badword_list();

    foreach ($blockedWords as $blockedWord) {
        if (strpos($code, $blockedWord) !== false) {
            return true;
        }
    }

    return false;
}

/**
 * generate a new random code
 */
function corona_test_results_ajax_get_unique_code($recursion = 0) {
    global $corona_test_results_cfg, $wpdb;

    // give up after 10000 retries to avoid endless recursion freezing the process
    if ($recursion >= 10000) {
        return false;
    }

    $randomCode = corona_test_results_get_random_code($corona_test_results_cfg['code_length']);

    if (corona_test_results_code_hasBadWord($randomCode) ) {
        return corona_test_results_ajax_get_unique_code($recursion+1);
    }

    $tableName = corona_test_results_get_table_name();
    $result = $wpdb->get_results ( $wpdb->prepare( "SELECT * FROM `$tableName` WHERE `code` = %s LIMIT 1", $randomCode) );

    if (count($result)) {
        return corona_test_results_ajax_get_unique_code($recursion+1);
    }

    return $randomCode;
}

/**
 * get a new random code via Ajax
 */
function corona_test_results_ajax_getcode() {
    global $wpdb;

    corona_test_results_conditionally_create_table();
    $tableName = corona_test_results_get_table_name();

	if ( isset( $_POST['batch'] ) ) {
		$batchCount = (int)$_POST['batch'];

		$result = array(
			'test_result_codes' => array()
		);

		if ($batchCount) {
			for ($n = 0; $n < $batchCount; $n++) {
				$newCode = corona_test_results_ajax_get_unique_code();
				if ($newCode) {
					$result['test_result_codes'][] = $newCode;
				}
			}

			if ( count($result['test_result_codes']) !== $batchCount ) {
				$result['error'] = 1;
				wp_send_json_error($result, 500);
				exit;
			}

			$insert_lists = array_fill(0, $batchCount, '(%s, %s)');
			$insert_lists_string = implode( ', ', $insert_lists );

			$values = array();
			$time_now = current_time('mysql', true);
			foreach ( $result['test_result_codes'] as $uniqueCode ) {
				$values[] = $uniqueCode;
				$values[] = $time_now;
			}

			$wpdb->query(
				$wpdb->prepare(
					"INSERT INTO `$tableName` (`code`,`created_at`) VALUES $insert_lists_string",
					$values
				)
			);

			$result['timestamp'] = $time_now;
			$result['status'] = __( 'Pending', 'corona-test-results' );
		}
	} else {
		$uniqueCode = corona_test_results_ajax_get_unique_code();
		$result = array(
			'test_result_code' => $uniqueCode
		);

		if (!$uniqueCode) {
			$result['error'] = 1;
			wp_send_json_error($result, 500);
			exit;
		}

		$wpdb->query(
			$wpdb->prepare(
				"INSERT INTO `$tableName` (`code`,`created_at`) VALUES (%s, %s) ",
				$uniqueCode,
				current_time('mysql', true)
			)
		);
	}

    print json_encode($result);
	wp_die();
}
add_action( 'wp_ajax_corona_test_results_ajax_getcode', 'corona_test_results_ajax_getcode' );

/**
 * get all rows for a given status
 */
function corona_test_results_fetch_rows_handler() {
	if ( isset( $_POST['stati'] ) && !empty( $_POST['stati'] ) && is_array( $_POST['stati'] ) ) {
		$stati = array();
		foreach( $_POST['stati'] as $id ) {
			$stati[] = intval( $id );
		}

		$options = corona_test_results_get_options();
		$entries_per_page = $options['assignation_entries_per_page'];

		$entries = corona_test_results_get_entries( $stati, $entries_per_page );

		if ( !empty( $entries ) ) {
			$entries = array_map(function( $row ) {
				$displayTime = date_i18n( get_option( 'date_format' ) . ', ' . get_option( 'time_format' ), strtotime(get_date_from_gmt($row->created_at, 'Y-m-d H:i:s')) );
				$timestamp = get_date_from_gmt($row->created_at, 'Y-m-d H:i:s');

				return array(
					'c' => $row->code,
					's' => (int)$row->status,
					't' => $timestamp,
					'd' => $displayTime
				);
			}, $entries);
		}

		wp_send_json( $entries );
	}
}
add_action( 'wp_ajax_corona_test_results_fetch_rows', 'corona_test_results_fetch_rows_handler' );

/**
 * dismissable notices
 */
function corona_test_results_dismiss_notice_handler() {
    $id = sanitize_html_class( $_POST['id'] );
    update_option( 'dismissed_corona_test_results_' . $id, TRUE );
}
add_action( 'wp_ajax_corona_test_results_dismiss_notice', 'corona_test_results_dismiss_notice_handler' );

/**
 * display settings link in plugin overview
 */
function corona_test_results_settings_link($links) {
	$settings_link = '<a href="admin.php?page=corona-test-results-settings">' . esc_html( vt_helper__default_i18n__( 'Settings' ) ) . '</a>';
	array_unshift($links, $settings_link);
	return $links;
}
add_filter("plugin_action_links_" . corona_test_results__mainfile( true ), 'corona_test_results_settings_link' );

/**
 * display Premium link in plugin overview
 */
function corona_test_results_premium_link( $links, $file ) {
	$options = corona_test_results_get_options();

    if ( corona_test_results__mainfile( true ) == $file ) {

		foreach ( $links as $index => $link ) {
			if ( strip_tags( $link ) === vt_helper__default_i18n__( 'Visit plugin site' ) ) {
				unset($links[$index]);
				break;
			}
		}

		if ( !isset( $options['license_key'] ) || empty( $options['license_key'] ) ) {
			$row_meta = array(
			'premium'    => '<a href="' . esc_url( corona_test_results_premium_shop_url() ) . '" target="_blank" style="color:#aa4800;">Get <strong style="font-style:italic">' . esc_html( 'Premium' ) . '</strong></a>'
			);

			return array_merge( $links, $row_meta );
		}

    }
    return (array) $links;
}
add_filter( 'plugin_row_meta', 'corona_test_results_premium_link', 10, 2 );

/**
 * display Support link in plugin overview
 */
function corona_test_results_support_link( $links, $file ) {
	$options = corona_test_results_get_options();

    if ( corona_test_results__mainfile( true ) == $file ) {

		$row_meta = array();

		if ( isset( $options['license_key'] ) && !empty( $options['license_key'] ) ) {
			$row_meta['support'] = corona_test_results_get_support_mail_link();
		}

		if ( corona_test_results_check_review_nag_timestamp() ) {
			$row_meta['rate_and_review'] = corona_test_results_get_review_link( '★★★★★ ' );
		}

		return array_merge( (array) $links, $row_meta );
    }

    return (array) $links;
}
add_filter( 'plugin_row_meta', 'corona_test_results_support_link', 10, 2 );

